import hypermedia.*;
import proxml.*;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

// Importing libraries needed for open/saving of files
import java.io.File;
import java.util.Hashtable;
import java.util.Enumeration;
import javax.swing.*;
import javax.swing.filechooser.*;


// Variables used to control the PlayViz main program
Timeline timeline;
Reel currentReel;
Frame currentFrame, nextFrame;
InfoBar infobar;
ToolBar toolbar;
Frame copiedFrame;
ArrayList copiedFrames = new ArrayList();
ArrayList exportFrames = new ArrayList();

// Variables used for control functions
boolean play, record;
int scrollpos, vscrollpos, timepos, oldtimepos;
float zoom;
float playtime, playstart, starttimepos;
boolean recorddata;
int playtimes = 0;

// Mouse control variables
boolean mouseReleased, rmousePressed, rmouseReleased, mouseDragged;
int mouseXDragged;

// Key control variables
boolean keyControl, keyShift, keyLeft, keyRight, keyUp, keyDown, keyC, keyV, keyX, keyZ;
//boolean key0, key1, key2, key3, key4, key5, key6, key7, key8, key9;

// Scrollbar variables
float SCROLL_MULTIPLIER = 9.33;
float VSCROLL_MULTIPLIER = 25;

// Font variable
PFont pvFont;

// UDP and XML stuff for Unreal connection
UDP udp;
final int UDP_SEND = 13777;
final int UDP_RECEIVE = 13888;

// XML variables for loading and saving PlayViz XML files
XMLInOut xmlIO;
XMLElement xmlTimeline;
XMLElement xmlReel;
XMLElement xmlFrame;

// Variables needed for file open/saving methods
JFileChooser chooser;
PVFileFilter fileFilter;

void setup()
{
  // Set up basic Processing variables
  size(900, 250);
  smooth();
  frameRate(15);

  // Load gui elements
  pvFont = loadFont("Verdana-10.vlw");
  textFont(pvFont);
  infobar = new InfoBar();
  toolbar = new ToolBar();

  // Initialize interface and UDP libraries
  // Arbitrarily choose a UDP port number
  udp = new UDP(this, UDP_RECEIVE);

  // Start listening for UDP stream
  udp.listen(true);
  udp.broadcast(true);

  timeline = new Timeline();
  timeline.addReel();
  currentReel = timeline.getReel(0);
  currentFrame = null;

  xmlIO = new XMLInOut(this);
  /*try
   {
   xmlIO.loadElement("test.xml");
   }
   catch(Exception e){ println("Could not find XML file"); }*/

  // Initialize PlayViz variables to their default values
  play = record = false;
  scrollpos = vscrollpos = timepos = 0;
  oldtimepos = -1;
  zoom = 2.0;
  playstart = -1;
  playtime = starttimepos = 0.0;

  // Initialize all the mouse variables to false to begin the program
  mouseReleased = rmousePressed = rmouseReleased = false;

  // Setup Java open/save dialog pop-ups
  chooser = new JFileChooser(new File("./data"));
  fileFilter = new PVFileFilter();
  fileFilter.addExtension("xml");
  fileFilter.setDescription("XML Files");
  chooser.setFileFilter(fileFilter);
}

// The draw function is equivalent to a "tick" in Processing, there
// should be approximately 30 ticks a second based upon the initial
// framerate setup
void draw()
{
  background(250);

  checkButtons(); // Check if buttons have been pressed each tick

  // If copy command is inputted, then copy the current frame into
  // the copyFrame variable to be stored until future use
  if(keyC)
  {
    if(currentFrame != null)
    {
      if(copiedFrames != null){ 
        copiedFrames.clear();
      }
      copiedFrames.add(currentFrame.copy());
      println(copiedFrames.size());
    }
  }
  //CHECK THIS CONDITION
  if(keyV)
  {
    int tempTime = timepos;
    if(currentReel != null && copiedFrames.size() > 0)
    {
      //trouble
      int i = 0;
      while(i < copiedFrames.size()){
        infobar.message("Copying frame at " + tempTime/zoom);
        float tempPos = tempTime/zoom;
        //if(copiedFrames.get(i) != null){
        Frame temp3 = ((Frame)copiedFrames.get(i)).copy();
        temp3.frameStart = tempPos;
        copiedFrames.set(i, temp3);
        //}  
        Frame temp4 = currentReel.getFrameAt(tempPos);
        if(copiedFrames.size() > 1){
          if(temp4 != null){
            /*if(i == (copiedFrames.size - 1)){
             copiedFrames = (Frame[])expand(copiedFrames, copiedFrames.length+1);
             }*/
            //copiedFrames.add(currentReel.getFrameAt(tempPos).copy());
            println("Did This");
          }
        }
        currentReel.addFrame((Frame)copiedFrames.get(i));
        if(((Frame)copiedFrames.get(i)).getLength() > 0.5){
          tempTime += ((Frame)copiedFrames.get(i)).getLength() + (((Frame)copiedFrames.get(i)).getLength() - 1);
        }
        tempTime++;
        i++;
      }
    }
  }
  if(keyX)
  {
    if(currentFrame != null)
    {
      currentFrame.interpolate = !currentFrame.interpolate;
    }
  }
  if(keyZ)
  {
    if(currentFrame != null)
    {
      currentFrame.show = !currentFrame.show;
    }
  }
  if(keyLeft && scrollpos > 0)
  {
    timeline.hs.spos -= SCROLL_MULTIPLIER;
    timeline.hs.newspos = timeline.hs.spos;
  }
  if(keyRight)
  {
    timeline.hs.spos += SCROLL_MULTIPLIER;
    timeline.hs.newspos = timeline.hs.spos;
  }

  if(keyUp && vscrollpos > 0)
  {
    timeline.vs.spos -= VSCROLL_MULTIPLIER;
    timeline.vs.newspos = timeline.vs.spos;
  }
  if(keyDown)
  {
    timeline.vs.spos += VSCROLL_MULTIPLIER;
    timeline.vs.newspos = timeline.vs.spos;
  }

  if(keyPressed && (key == '0')){
    if(timeline.numReels() >= 1){
      currentReel = timeline.getReel(0);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  } 

  if(keyPressed && (key == '1')){
    if(timeline.numReels() >= 2){
      currentReel = timeline.getReel(1);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '2')){
    if(timeline.numReels() >= 3){
      currentReel = timeline.getReel(2);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '3')){
    if(timeline.numReels() >= 4){
      currentReel = timeline.getReel(3);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '4')){
    if(timeline.numReels() >= 5){
      currentReel = timeline.getReel(4);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '5')){
    if(timeline.numReels() >= 6){
      currentReel = timeline.getReel(5);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '6')){
    if(timeline.numReels() >= 7){
      currentReel = timeline.getReel(6);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '7')){
    if(timeline.numReels() >= 8){
      currentReel = timeline.getReel(7);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '8')){
    if(timeline.numReels() >= 9){
      currentReel = timeline.getReel(8);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  if(keyPressed && (key == '9')){
    if(timeline.numReels() >= 10){
      currentReel = timeline.getReel(9);
      if(currentReel.hasFrames()){
        currentFrame = currentReel.getFrame(0);
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);      
        //play = false;
        //record = false;
      }
    }
  }

  toolbar.update();

  if(timeline != null)
  {
    // Explaining how the scroll bar works
    // The scrollbar value from getPos() gives the basic pixel position of the scrollbar.
    // Thus since we've set up the reelname overflow to be about 100 and the center of the
    // horizontal scroll button is about 2 so we subtract 102 to get to around 0 rounded.
    // The max value then is whatever the length of the program is, in this case 900, but 
    // with the subtracted constant value we get approximately between 0 - 800. We then divide
    // this amount by sum constant and then change cast it to an int to get our scrollpos
    // variable. The same principle applies for the vertical scrollbar.

    int adjustScrollpos = int((timeline.maxTime - (width - Reel.REELNAME_OVERFLOW - 15)/(15*zoom))*zoom) + 3;
    if(adjustScrollpos < 1)
    {
      adjustScrollpos = 1;
    }
    SCROLL_MULTIPLIER = (width - Reel.REELNAME_OVERFLOW)/adjustScrollpos;

    // Update position based on scrollbar
    scrollpos = (int)((timeline.hs.getPos() - Reel.REELNAME_OVERFLOW - 2)/SCROLL_MULTIPLIER); // Ranges from 0 - 75
    vscrollpos = abs((int)((timeline.vs.getPos() + 67)/20));

    //println(timeline.hs.getPos() + " " + scrollpos);
    //println(timeline.vs.getPos() + " " + vscrollpos);

    timeline.update();
    timeline.draw();
  }

  infobar.draw();
  toolbar.draw();

  if(udp.isClosed())
  {
    println("Cannot receive messages");
  }

  // If the current mode is play then we need to start automatically
  // incrementing the time position marker
  if(play && timeline != null)
  {
    if(playstart < 0)
    {
      playstart = millis()/1000.0;
      starttimepos = timepos/zoom;
    }
    playtime = millis()/1000.0 - playstart + starttimepos;

    if(currentFrame != null)
    {
      if(currentFrame.show && currentFrame.interpolate &&
        playtime >= currentFrame.showStart() + currentFrame.getStartTime() &&
        playtime < currentFrame.showStart() + currentFrame.showLength() + currentFrame.getStartTime())
      {
        if (nextFrame != null)
        {

          //___________________________________________________________________________________________________________________________________
          //Grab the values of the next frame positions/rotations.
          //x = vertical rotation (problematic)
          //y = horizontal rotation
          float x1, x2, y1, y2;
          float pos1, pos2, pos3;

          x1 = currentFrame.getRotation().x();
          y1 = currentFrame.getRotation().y();
          x2 = nextFrame.getRotation().x();
          y2 = nextFrame.getRotation().y();

          pos1 = currentFrame.getPosition().x();
          pos2 = currentFrame.getPosition().y();
          pos3 = currentFrame.getPosition().z();

          float dxp, dyp, dzp, dxr, dyr;

          dxp = (nextFrame.getPosition().x() - currentFrame.getPosition().x())/currentFrame.getLength()*(playtime - currentFrame.getStartTime());
          dyp = (nextFrame.getPosition().y() - currentFrame.getPosition().y())/currentFrame.getLength()*(playtime - currentFrame.getStartTime());
          dzp = (nextFrame.getPosition().z() - currentFrame.getPosition().z())/currentFrame.getLength()*(playtime - currentFrame.getStartTime());
          dxr = (nextFrame.getRotation().x() - currentFrame.getRotation().x())/currentFrame.getLength()*(playtime - currentFrame.getStartTime());
          dyr = (nextFrame.getRotation().y() - currentFrame.getRotation().y())/currentFrame.getLength()*(playtime - currentFrame.getStartTime());


          /*if( ){
           x2 = nextFrame.getRotation().x();
           
           if(x2 >= 65536){
           x2 = x2 - 65535;
           }             
           else if(x2 <= 0){
           x2 = 65535 + x2;
           }
           //currentFrame = currentReel.getNextFrame(currentFrame);
           }*/
          if( (x2 - x1) >= 30000 || (x2 - x1) <=-30000 ||
            (y2 - y1) >= 30000 || (y2 - y1) <=-30000) {
            int i = 0;
            while(i < 5){             
              pos1 = currentFrame.getPosition().x() + dxp/5;
              pos2 = currentFrame.getPosition().y() + dyp/5;
              pos3 = currentFrame.getPosition().z() + dzp/5;
              y2 = currentFrame.getRotation().y() + dyr/5;
              x2 = currentFrame.getRotation().x() + dxr/5;
              //x2 += dxr/5;

              if(x2 >= 65536){
                x2 = 1;
                String interpolateString = (pos1 + "," + pos2 + "," + pos3 + " " + x2 + "," + y2 + "," + 0); 
                x2 = nextFrame.getPosition().x() / 2;
              }             
              else if(x2 <= 0){
                x2 = 65535;
                String interpolateString = (pos1 + "," + pos2 + "," + pos3 + " " + x2 + "," + y2 + "," + 0); 
                x2 =  x2 + ((nextFrame.getPosition().x() + 65535) / 2);
              }
              if(y2 >= 65536){
                y2 = y2 - 65535;
              }             
              else if(y2 <= 0){
                y2 = 65535 + y2;
              }

              String interpolateString = (pos1 + "," + pos2 + "," + pos3 + " " + x2 + "," + y2 + "," + 0); 

              udp.send(interpolateString, "localhost", UDP_SEND);
              i++;
              int j = 0;
              //while(j<1000000){j++;}
            }
          }

          else {
            String interpolateString = ((nextFrame.getPosition().x() - currentFrame.getPosition().x())/currentFrame.getLength()*(playtime - currentFrame.getStartTime()) + currentFrame.getPosition().x()) + "," +
              ((nextFrame.getPosition().y() - currentFrame.getPosition().y())/currentFrame.getLength()*(playtime - currentFrame.getStartTime()) + currentFrame.getPosition().y()) + "," +
              ((nextFrame.getPosition().z() - currentFrame.getPosition().z())/currentFrame.getLength()*(playtime - currentFrame.getStartTime()) + currentFrame.getPosition().z()) + " " +
              ((nextFrame.getRotation().x() - currentFrame.getRotation().x())/currentFrame.getLength()*(playtime - currentFrame.getStartTime()) + currentFrame.getRotation().x()) + "," +
              ((nextFrame.getRotation().y() - currentFrame.getRotation().y())/currentFrame.getLength()*(playtime - currentFrame.getStartTime()) + currentFrame.getRotation().y()) + "," + 0;
            udp.send(interpolateString, "localhost", UDP_SEND);         
          }

          //println("Sent: " + currentFrame.getString());
          //  currentFrame = nextFrame;
          // nextFrame = currentReel.getNextFrame(currentFrame);
          //___________________________________________________________________________________________________________________________________
        }
        else
        {
          println("Sent: " + currentFrame.getString());
          udp.send(currentFrame.getString(), "localhost", UDP_SEND);
          exportFrames.add(currentFrame.copy());
          }
        }
      else if(currentFrame.show &&
        playtime >= currentFrame.showStart() + currentFrame.getStartTime() &&
        playtime < currentFrame.showStart() + currentFrame.showLength() + currentFrame.getStartTime())
      {
        println("Sent: " + currentFrame.getString());
        udp.send(currentFrame.getString(), "localhost", UDP_SEND);
        exportFrames.add(currentFrame.copy());
        }
      else if(playtime < currentFrame.showStart() + currentFrame.getStartTime() ||
        playtime >= currentFrame.showStart() + currentFrame.showLength() + currentFrame.getStartTime())
      {
        // Find new current frame
        for(int i = 0; i < timeline.numReels(); i++)
        {
          Reel tempReel = timeline.getReel(i);
          currentFrame = tempReel.getFrameAt(playtime);
          if(currentFrame != null)
          {
            if(currentFrame.show && 
              playtime >= currentFrame.showStart() + currentFrame.getStartTime() &&
              playtime < currentFrame.showStart() + currentFrame.showLength() + currentFrame.getStartTime())
            {
              println("Sent: " + currentFrame.getString());
              udp.send(currentFrame.getString(), "localhost", UDP_SEND);
              exportFrames.add(currentFrame.copy());
                currentReel = tempReel;
              break;
            }
            currentFrame = null;
          }
        }

        if(currentFrame != null)
        {
          if(currentFrame.interpolate)
          {
            nextFrame = currentReel.getNextFrame(currentFrame);
          }
        }
      }
    }

    //here
    else
    {
      for(int i = 0; i < timeline.numReels(); i++)
      {
        Reel tempReel = timeline.getReel(i);
        currentFrame = tempReel.getFrameAt(playtime);
        if(currentFrame != null)
        {
          if(currentFrame.show && 
            playtime >= currentFrame.showStart() + currentFrame.getStartTime() &&
            playtime < currentFrame.showStart() + currentFrame.showLength() + currentFrame.getStartTime())
          {
            println("Sent: " + currentFrame.getString());
            exportFrames.add(currentFrame.copy());
              udp.send(currentFrame.getString(), "localhost", UDP_SEND);
            break;
          }
          currentFrame = null;          
        }
      }

      if(currentFrame != null)
      {
        if(currentFrame.interpolate)
        {
          nextFrame = currentReel.getNextFrame(currentFrame);
        }
      }
    }

    timepos = (int)(playtime*zoom);

    // If the time position marker goes off the screen, automatically
    // scroll a whole screen forward so the marker is always on-screen
    if(timepos - scrollpos > 44)
    {
      timeline.hs.spos += 44*SCROLL_MULTIPLIER;
      timeline.hs.newspos = timeline.hs.spos;
    }
  }
  // If record is true, also do some basic timekeeping
  else if(record)
  {
    if(playstart < 0)
    {
      playstart = millis()/1000.0;
      starttimepos = timepos/zoom;
      recorddata = true;
    }
    playtime = millis()/1000.0 - playstart + starttimepos;

    if(timepos != (int)playtime*zoom)
    {
      recorddata = true;
    }

    timepos = (int)(playtime*zoom);

    // If the time position marker goes off the screen, automatically
    // scroll a whole screen forward so the marker is always on-screen
    if(timepos - scrollpos > 44)
    {
      timeline.hs.spos += 44*SCROLL_MULTIPLIER;
      timeline.hs.newspos = timeline.hs.spos;
    }
  }

  // Reset all the necessary control variables at the end of each tick,
  // including all the mouse and key variables not handled by Processing.
  // The CTRL boolean variable is not resetted because we want to recognize
  // when it is pressed and NOT when it is released.
  if(!mousePressed && !rmousePressed)
  {
    mouseDragged = false;
    mouseXDragged = 0;
  }

  if(mouseDragged)
  {
    mouseXDragged += mouseX - pmouseX;
  }

  // This handles the dragging interface for changing the shown portion of
  // a frame. It calculates where the relative start and end points of the
  // highlighted portions are, takes all the frames selected within that scope,
  // and adjusts the shown portion accordingly.

  //CHECK HERE AT KEY UNTIL copiedFrames = (Frame[])temp;
  if(rmouseReleased && mouseXDragged > 5)
  {
    if(currentReel != null)
    {
      int pos1 = oldtimepos;
      int pos2 = scrollpos + (mouseX - currentReel.REELNAME_OVERFLOW)/15;

      List frames = currentReel.getFramesAt(pos1, pos2);
      //println(pos1/zoom + " " + pos2/zoom);

      for(int i = 0; i < frames.size(); i++)
      {
        Frame f = (Frame)frames.get(i);

        if(!keyControl && !keyShift)
        {
          f.show = true;
        }

        if(pos1/zoom != f.getStartTime() + f.getLength() && pos2/zoom != f.getStartTime())
        {
          if(keyShift)
          {
            currentReel.removeFrame(f);
          }
          else if(keyControl)
          {
            f.show = !f.show;
          }
          else if(key == 'a' || key == 'A')
          {
            if(i == 1){ 
              copiedFrames.clear();
            }
            copiedFrames.add(f.copy());

          }

          else if(key == 's' || key == 'S')
          {
            if(i == 1){ 
              copiedFrames.clear();
            }
            copiedFrames.add(f.copy());
            currentReel.removeFrame(f);
          }

          else if(key == 'd' || key == 'D')
          {
            if(f.frameStart >= pos1/zoom)
            {
              f.showStart = 0.0;
            }
            else
            {
              f.showStart = pos1/zoom - f.getStartTime();
            }

            //println(f.showStart());

            if(f.getStartTime() + f.getLength() <= pos2/zoom)
            {
              f.showLength = f.getLength() - f.showStart();
            }
            else
            {
              f.showLength = pos2/zoom - f.getStartTime() - f.showStart();
            }
          }
        }
      }
      infobar.message("Copy Size: " + copiedFrames.size());
      //copiedFrames = (Frame[])temp;
      oldtimepos = -1;
    }
  }

  mouseReleased = false;
  rmouseReleased = false;
  keyC = keyV = keyX = keyZ = false;

  // If none of the mouse buttons are being pressed then logically the
  // right mouse button is not pressed
  if(!mousePressed)
  {
    rmousePressed = false;
  }
}

void receive(byte[] data)
{
  //data = contract(data, data.length - 2);
  String message = new String(data);

  println(message);

  if(record && currentReel != null)
  {
    if(message.equals("end_socket"))
    {
    }
    else if(recorddata)
    {
      StringTokenizer st = new StringTokenizer(message);

      if(st.countTokens() == 2)
      {
        StringTokenizer sl = new StringTokenizer(st.nextToken(), ",");
        StringTokenizer sr = new StringTokenizer(st.nextToken(), ",");

        if(sl.countTokens() == 3 && sr.countTokens() == 3)
        {
          // Assign the position and rotation vectors to some temporary variables
          float posX = Float.valueOf(sl.nextToken()).floatValue();
          float posY = Float.valueOf(sl.nextToken()).floatValue();
          float posZ = Float.valueOf(sl.nextToken()).floatValue();
          float rotX = Float.valueOf(sr.nextToken()).floatValue();
          float rotY = Float.valueOf(sr.nextToken()).floatValue();
          float rotZ = Float.valueOf(sr.nextToken()).floatValue();

          //println(posX + "," + posY + "," + posZ + " " + rotX + "," + rotY + "," + rotZ);

          // If the current frame is not null, then we should check for same basic things
          if(currentFrame != null)
          {

            println(currentFrame.getPosition().x() + " " +
              currentFrame.getPosition().y() + " " +
              currentFrame.getPosition().z() + " " +
              currentFrame.getRotation().x() + " " +
              currentFrame.getRotation().y() + " " +
              currentFrame.getRotation().z());

            if(currentFrame.getPosition().x() == posX &&
              currentFrame.getPosition().y() == posY &&
              currentFrame.getPosition().z() == posZ &&
              currentFrame.getRotation().x() == rotX &&
              currentFrame.getRotation().y() == rotY &&
              currentFrame.getRotation().z() == rotZ)
            {
              currentFrame.frameLength = timepos/zoom - currentFrame.frameStart + 1.0/zoom;
              currentFrame.showLength = currentFrame.frameLength;
            }
            else if(currentFrame.frameStart != timepos/zoom)
            {
              currentReel.addFrame(posX, posY, posZ, rotX, rotY, rotZ, 1.0/zoom);
            }
          }
          // Otherwise (i.e. this is the first frame added)
          else
          {
            currentReel.addFrame(posX, posY, posZ, rotX, rotY, rotZ, 1.0/zoom);
          }
        }
      }
      recorddata = false;
    }
  }
}

void checkButtons()
{
  // Button presses
  if(toolbar.recordbutton.pressed)
  {
    record = true;

    toolbar.recordbutton.locked = true;

    currentFrame = null;

    infobar.message("Recording started");
  }

  if(toolbar.playbutton.pressed)
  {
    play = true;

    toolbar.playbutton.locked = true;

    infobar.message("Playing recording");
    exportFrames.clear();
  }

  if(toolbar.stopbutton.pressed)
  {
    if(record)
    {
      infobar.message("Recording halted");
    }
    else if(play)
    {
      infobar.message("Playback halted");
      
    }

    toolbar.recordbutton.locked = false;
    toolbar.playbutton.locked = false;
    toolbar.stopbutton.locked = false;

    // Reset play variables
    playtime = 0.0;
    playstart = -1.0;

    record = false;
    play = false;
  }

  if(toolbar.newbutton.pressed)
  {
    timeline = new Timeline();
    timeline.addReel();
    currentReel = timeline.getCurrentReel();

    infobar.message("New timeline created!");
  }

  if(toolbar.savebutton.pressed)
  {
    int returnVal = chooser.showSaveDialog(this);
    if(returnVal == JFileChooser.APPROVE_OPTION)
    {
      File savedFile = chooser.getSelectedFile();
      buildXMLTree();
      try
      {
        xmlIO.saveElement(xmlTimeline, savedFile.getName() + ".xml");

        infobar.message("File saved as " + savedFile.getName() + ".xml");
      }
      catch(Exception e) { 
        println("Could not save file"); 
      }
    }
  }

  if(toolbar.exportbutton.pressed)
  {
    playToMaya();
  }

  if(toolbar.loadbutton.pressed)
  {
    int returnVal = chooser.showOpenDialog(this);
    if(returnVal == JFileChooser.APPROVE_OPTION)
    {
      try
      {
        //println(chooser.getSelectedFile().getName());
        xmlIO.loadElement(chooser.getSelectedFile().getName());

        infobar.message(chooser.getSelectedFile().getName() + " successfully opened");
      }
      catch(Exception e) { 
        println("Could not find XML file"); 
      }
    }
  }

  if(toolbar.addbutton.pressed)
  {
    timeline.addReel();
    currentReel = timeline.getCurrentReel();

    infobar.message("New reel added!");
  }

  if(toolbar.delbutton.pressed)
  {
    if(currentReel != null)
    {
      timeline.removeReel(currentReel);

      infobar.message(currentReel.name + " deleted from timeline"); 

      currentReel = null;
    }
  }

  if(toolbar.zoominbutton.pressed)
  {
    if(zoom < 32)
    {
      zoom *= 2;
      infobar.message("Zoomed in");
    }
    else
    {
      infobar.message("Cannot zoom in further");
    }    
  }

  if(toolbar.zoomoutbutton.pressed)
  {
    zoom /= 2;

    infobar.message("Zoomed out");
  }
}

void mousePressed()
{
  if(mouseButton == RIGHT)
  {
    rmousePressed = true;
  }
}

void mouseReleased()
{
  if(mouseButton == LEFT)
  {
    mouseReleased = true;
  }
  else if(mouseButton == RIGHT)
  {
    rmouseReleased = true;
  }

  if(mouseButton == LEFT && currentReel != null)
  {
    if(currentReel.over)
    {
      currentReel.pressed = true;
      currentReel.t_pressed = 0;
    }
  }
}

void mouseDragged()
{
  mouseDragged = true;
}

// Handle boolean key control variables not built into processing
void keyPressed()
{
  if(key == CODED)
  {
    if(keyCode == CONTROL)
    {
      keyControl = true;
    }
    if(keyCode == SHIFT)
    {
      keyShift = true;
    }
    if(keyCode == LEFT)
    {
      keyLeft = true;
    }
    if(keyCode == RIGHT)
    {
      keyRight = true;
    }
    if(keyCode == UP)
    {
      keyUp = true;
    }
    if(keyCode == DOWN)
    {
      keyDown = true;
    }
  }
}

void keyReleased()
{
  if(key == CODED)
  {
    if(keyCode == CONTROL)
    {
      keyControl = false;
    }
    if(keyCode == SHIFT)
    {
      keyShift = false;
    }
    if(keyCode == LEFT)
    {
      keyLeft = false;
    }
    if(keyCode == RIGHT)
    {
      keyRight = false;
    }
    if(keyCode == UP)
    {
      keyUp = false;
    }
    if(keyCode == DOWN)
    {
      keyDown = false;
    }
  }
  // In PlayViz the delete command should delete the current camera
  // movement occupying a time slice, thus if the currentReel is not
  // null and the currentFrame is not null then remove the currentFrame
  // from the currentReel (i.e. delete it)
  else if(key == DELETE)
  {
    if(currentReel != null && currentFrame != null)
    {
      currentReel.frames.remove(currentFrame);
    }
  }
  else if(key == 'Z' || key == 'z')
  {
    keyZ = true;
  }
  else if(key == 'X' || key == 'x')
  {
    keyX = true;
  }
  else if(key == 'C' || key == 'c')
  {
    keyC = true;
  }
  else if(key == 'V' || key == 'v')
  {
    keyV = true;
  }
}
