class Reel
{
  List frames;
  String name;

  int position;
  boolean over, selected, pressed;
  boolean overFrameStart, overFrameEnd, overFrameStartLock, overFrameEndLock;
  int t_pressed;

  static final int REELNAME_OVERFLOW = 100;

  public Reel(int position)
  {
    frames = new ArrayList();

    name = "Camera" + position;

    this.position = position;
  }

  // Function for adding a frame to a reel where start time is based upon
  // the current time position on the reel
  public boolean addFrame(double xpos, double ypos, double zpos,
  double xrot, double yrot, double zrot,
  float frameLength)
  {
    return this.addFrame(xpos, ypos, zpos, xrot, yrot, zrot, timepos/zoom, frameLength, true, true, 0.0, frameLength);
  }

  // Function for adding a frame to a reel when the frame already has a set
  // start time position
  public boolean addFrame(double xpos, double ypos, double zpos,
  double xrot, double yrot, double zrot,
  float frameStart, float frameLength, boolean interpolate,
  boolean show, float showStart, float showLength)
  {
    return this.addFrame(new Frame(new Vector(xpos, ypos, zpos), new Vector(xrot, yrot, zrot), frameStart, frameLength, interpolate, show, showStart, showLength));    
  }

  public boolean addFrame(Frame f)
  {
    currentFrame = f;

    timeline.calculateMaxTime(f);

    return frames.add(currentFrame);
  }

  // Gives the current y position in pixels for the frame, used specifically
  // since this a number commonly used throughout the draw function
  private int ypos() { 
    return 60 + (position - vscrollpos)*30; 
  }

  public boolean hasFrames() { 
    if(frames.size() > 0)
      return true; 
    else
      return false;
  }
  public int numFrames() { 
    return frames.size(); 
  }
  public Frame getFrame(int i) { 
    return (Frame)frames.get(i); 
  }
  public boolean removeFrame(Frame f) { 
    return frames.remove(f); 
  }

  void update()
  {
    for(int i = 0; i < frames.size(); i++)
    {
      Frame f = getFrame(i);

      if((f.frameStart*zoom >= scrollpos && f.frameStart*zoom <= scrollpos + 25*zoom) ||
        ((f.frameStart + f.frameLength)*zoom >= scrollpos && (f.frameStart + f.frameLength)*zoom <= scrollpos + 25*zoom))
      {
        f.update(REELNAME_OVERFLOW, ypos());
      }
    }

    if(currentReel == this)
    {
      selected = true;
    }  
    else
    {
      selected = false;
    }

    if(over())
    {
      over = true;
    }
    else
    {
      over = false;
    }

    if(mousePressed && over && !pressed && !mouseDragged)
    {
      currentReel = this;
      fill(0,255,255);
    }
    /*else if(mousePressed && over && !pressed && currentReel != this && currentFrame != null)
     {
     currentReel.removeFrame(currentFrame);
     currentReel = this;
     currentReel.addFrame(currentFrame);
     }*/
    else if(rmousePressed && over && oldtimepos < 0)
    {
      oldtimepos = scrollpos + (int)((mouseX - REELNAME_OVERFLOW)/15);
    }
    // If a doubleclick event occurs then if the time slice has some portion of a
    // camera movement (frame) involved then it should trasmit that frame through
    // the UDP protocol
    else if (mousePressed && over && pressed)
    {
      // Doubleclick event
      //currentFrame = getClickedFrame();
      if(currentFrame != null)
      {
        println("Sent: " + currentFrame.getString());

        udp.send(currentFrame.getString(), "localhost", UDP_SEND);
      }

      pressed = false;
    }

    // If the mouse is being dragged by the user and the user is "in control" of the start
    // of a frame (i.e. camera control) then the frame's start time needs to be modified.
    if(overFrameStartLock)
    {
      currentFrame.frameStart = timepos/zoom;
    }
    if(overFrameEndLock)
    {
      if(timepos/zoom - currentFrame.frameStart > 0)
      {
        currentFrame.frameLength = timepos/zoom - currentFrame.frameStart;
      }
    }

    if(mousePressed && overFrameStart && !overFrameStartLock)
    {
      overFrameStartLock = true;
    }
    if(mousePressed && overFrameEnd && !overFrameEndLock)
    {
      overFrameEndLock = true;
    }

    if(!mousePressed || play)
    {
      overFrameStartLock = overFrameEndLock = false;
    }

    if(pressed)
    {
      t_pressed++;
      if(t_pressed > 5)
      {
        pressed = false;
      }
    }
  }

  public List getFramesAt(int pos1, int pos2)
  {
    List f = new ArrayList();
    for(int i = 0; i < numFrames(); i++)
    {
      Frame tempFrame = getFrame(i);
      if(tempFrame.getStartTime()*zoom <= pos2 && (tempFrame.getStartTime() + tempFrame.getLength())*zoom >= pos1)
      {
        f.add(tempFrame);
      }
    }
    return f;
  }

  public List getFramesAt(int pos)
  {
    return this.getFramesAt(pos, pos);
  }

  public Frame getFrameAt(int pos)
  {
    List tempFrames = this.getFramesAt(pos);
    if(tempFrames.size() > 0)
    {
      return (Frame)(tempFrames.remove(tempFrames.size() - 1));
    }
    else
    {
      return null;
    }
  }

  public Frame getFrameAt(float time)
  {
    for(int i = 0; i < numFrames(); i++)
    {
      Frame tempFrame = getFrame(i);
      if(tempFrame.getStartTime() <= time && tempFrame.getStartTime() + tempFrame.getLength() >= time)
      {
        return tempFrame;
      }
    }
    return null;
  }

  public Frame getNextFrame(Frame f)
  {
    Frame next = null;

    for(int i = 0; i < frames.size(); i++)
    {
      Frame temp = getFrame(i);
      if(next != null)
      {
        if(temp.getStartTime() > f.getStartTime() && temp.getStartTime() < next.getStartTime())
        {
          next = temp;
        }
      }
      else if(temp.getStartTime() > f.getStartTime())
      {
        next = temp;
      }
    }

    return next;
  }

  public Frame getPreviousFrame(Frame f)
  {
    Frame previous = null;

    for(int i = 0; i < frames.size(); i++)
    {
      Frame temp = getFrame(i);
      if(previous != null)
      {
        if(temp.getStartTime() < f.getStartTime() && temp.getStartTime() > previous.getStartTime())
        {
          previous = temp;
        }
      }
      else if(temp.getStartTime() < f.getStartTime())
      {
        previous = temp;
      }
    }

    return previous;
  }

  private boolean over()
  {
    if(mouseX > 0 && mouseX < width - 15 &&
      mouseY > ypos() && mouseY < ypos() + 30)
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  public void draw()
  {
    // Timebar
    if(position % 2 == 1)
    {
      noStroke();
      fill(250, 240, 240);
      rect(100, ypos(), width - REELNAME_OVERFLOW, 30);
    }

    // Draw time slices in each reel
    for(int i = scrollpos; i < (int)(width - REELNAME_OVERFLOW)/15 + scrollpos; i++)
    {
      // Darkened bars every 5 seconds
      if(i % (5*zoom) == 0)
      {
        if(position % 2 == 0)
        {
          fill(235);
        }
        else
        {
          fill(240, 230, 230);
        }
        rect(REELNAME_OVERFLOW + (i - scrollpos)*15, ypos(), 15, 30);
      }
      stroke(225);
      line(REELNAME_OVERFLOW + (i - scrollpos)*15, ypos(), REELNAME_OVERFLOW + (i - scrollpos)*15, ypos() + 30);
    }  

    // If mouse is being dragged, highlight
    if((mouseButton == LEFT) && mousePressed && over){
      fill(0, 255, 255);
    }

    if((mouseButton == RIGHT) && mouseDragged && over)
    {
      if(keyControl) // Change control if CTRL held to indicate different kind of selection
      {
        fill(255, 185, 15);
      }
      else if(keyShift)
      {
        fill(255, 0, 0);
      }
      else if(keyPressed && (key == 'a' || key == 'A'))
      {
        fill(0,0,255);
      }
      else if(keyPressed && (key == 's' || key == 'S'))
      {
        fill(255, 255, 0);
      }

      else if(keyPressed && (key == 'd' || key == 'D'))
      {
        fill(255, 255, 0);
      }
      for(int i = oldtimepos; i <= scrollpos + (mouseX - REELNAME_OVERFLOW)/15; i++)
      {
        rect(REELNAME_OVERFLOW + (i - scrollpos)*15, ypos(), 15, 30);
      }
    }


    stroke(0);
    line(REELNAME_OVERFLOW, ypos(), width, ypos());
    line(REELNAME_OVERFLOW, ypos() + 30, width, ypos() + 30);

    // Draw frames here
    // Draw circles indicating frames
    float time_elapsed = 0.0;    

    for(int i = 0; i < frames.size(); i++)
    {
      Frame f = getFrame(i);

      if((f.frameStart*zoom >= scrollpos && f.frameStart*zoom <= scrollpos + 25*zoom) ||
        ((f.frameStart + f.frameLength)*zoom >= scrollpos && (f.frameStart + f.frameLength)*zoom <= scrollpos + 25*zoom))
      {
        f.draw(REELNAME_OVERFLOW, ypos());
      }
    }

    // Name
    stroke(200);
    line(0, ypos(), REELNAME_OVERFLOW, ypos());
    line(0, ypos() + 30, REELNAME_OVERFLOW, ypos() + 30);
    if(selected)
    {
      fill(200);
    }
    else
    {
      fill(255);
    }
    rect(0, ypos(), REELNAME_OVERFLOW, 30);
    fill(0);
    text(name, 10, ypos() + 20);
  }
}

class Frame
{
  Vector position, rotation;
  float frameStart, frameLength;
  boolean interpolate;
  boolean show;
  float showStart, showLength;
  boolean over, overStart, overEnd, startLock, endLock;

  public Frame(Vector position, Vector rotation, float frameStart, float frameLength, boolean interpolate, boolean show, float showStart, float showLength)
  {
    this.position = position;
    this.rotation = rotation;
    this.frameStart = frameStart;
    this.frameLength = frameLength;
    this.interpolate = interpolate;
    this.show = show;
    this.showStart = showStart;
    this.showLength = showLength;
    this.over = false;
  }

  public float getStartTime() { 
    return this.frameStart; 
  }
  public float getLength() { 
    return this.frameLength; 
  }
  public float showStart() { 
    return this.showStart; 
  }
  public float showLength() { 
    return this.showLength; 
  }
  public Vector getPosition() { 
    return this.position; 
  }
  public Vector getRotation() { 
    return this.rotation; 
  }

  public String getString()
  {
    return (position.x + "," + position.y + "," + position.z + " " + rotation.x + "," + rotation.y + "," + rotation.z);
  }

  public Frame copy()
  {
    return new Frame(this.position.copy(), this.rotation.copy(), this.frameStart, this.frameLength, this.interpolate, this.show, this.showStart, this.showLength);
  }

  void update(int x, int y)
  {
    this.over(x, y);

    if(overStart && mousePressed && !rmousePressed && !mouseDragged)
    {
      startLock = true;
    }
    if(overEnd && mousePressed && !rmousePressed && !mouseDragged)
    {
      endLock = true;
    }

    if(startLock)
    {
      if((timepos + scrollpos)/zoom >= 0)
      {
        frameStart = (timepos + scrollpos)/zoom;
      }

      /*Frame previous = currentReel.getPreviousFrame(this);
       if(previous != null)
       {
       if(previous.getStartTime() + previous.getLength() > frameStart)
       {
       frameStart = previous.getStartTime() + previous.getLength();
       }
       }*/
    }
    if(endLock)
    {
      if((timepos + scrollpos)/zoom - frameStart >= 1.0/zoom)
      {
        frameLength = (timepos + scrollpos)/zoom - frameStart;

        Frame next = currentReel.getNextFrame(this);
        if(next != null)
        {
          if(next.getStartTime() < frameStart + frameLength)
          {
            next.frameStart = frameStart + frameLength;
          }
        }
      }
      else
      {
        frameLength = 1.0/zoom;
      }

      if(showLength > frameLength)
      {
        showLength = frameLength;
      }
    }

    if(!mousePressed)
    {
      startLock = false;
      endLock = false;
    }
  }

  public boolean over(int x, int y)
  {
    if(mouseX > x + 2 + (frameStart*zoom + scrollpos)*15 && mouseX < x + 13 + (frameStart*zoom + scrollpos)*15 &&
      mouseY > y + 5 && mouseY < y + 16)
    {
      overStart = true;
    }
    else
    {
      overStart = false;
    }

    if(mouseX > x + 2 + ((frameStart + frameLength)*zoom + scrollpos)*15 && mouseX < x + 13 + ((frameStart + frameLength)*zoom + scrollpos)*15 &&
      mouseY > y + 17 && mouseY < y + 28)
    {
      overEnd = true;
    }
    else
    {
      overEnd = false;
    }

    return false;
  }

  public void draw(int x, int y)
  {    
    // Draw the orange outline indicating "shown" portion
    if(this.show)
    {
      noStroke();
      fill(255, 170, 10);
      if(showStart < frameLength/2)
      {
        if(showLength + showStart >= frameLength/2)
        {
          rect(x + 8 + ((showStart + frameStart)*zoom - scrollpos)*15, y + 9, (frameLength/2 - showStart)*zoom*15, 3);
        }
        else
        {
          rect(x + 8 + ((showStart + frameStart)*zoom - scrollpos)*15, y + 9, (showLength)*zoom*15, 3);
        }
      }
      if(showStart <= frameLength/2 && showLength + showStart >= frameLength/2)
      {
        rect(x + 7 + ((frameStart + frameLength/2)*zoom - scrollpos)*15, y + 9, 3, 13);
      }
      if(showLength + showStart > frameLength/2)
      {
        if(showStart < frameLength/2)
        {
          rect(x + 8 + ((frameStart + frameLength/2)*zoom - scrollpos)*15, y + 20, (showLength + showStart - frameLength/2)*zoom*15, 3);
        }
        else
        {
          rect(x + 8 + ((frameStart + showStart)*zoom - scrollpos)*15, y + 20, (showLength)*zoom*15, 3);
        }
      }
    }

    // Connecting lines
    strokeWeight(1);
    stroke(175);
    line(x + 8 + (frameStart*zoom - scrollpos)*15, y + 11, x + 8 + ((frameStart + frameLength/2)*zoom - scrollpos)*15, y + 11);
    line(x + 8 + ((frameStart + frameLength/2)*zoom - scrollpos)*15, y + 11, x + 8 + ((frameStart + frameLength/2)*zoom - scrollpos)*15, y + 22);
    line(x + 8 + ((frameStart + frameLength/2)*zoom - scrollpos)*15, y + 22, x + 2 + ((frameStart + frameLength)*zoom - scrollpos)*15, y + 22);

    // Circles
    strokeWeight(1);
    if(show && showStart == 0.0)
    {
      fill(255, 170, 10);
      stroke(255, 170, 10);
      ellipse(x + 8 + (frameStart*zoom - scrollpos)*15, y + 10, 11, 11);
    }
    fill(0);
    stroke(0);
    ellipse(x + 8 + (frameStart*zoom - scrollpos)*15, y + 10, 9, 9);
    noFill();
    if(show && showLength - showStart == frameLength)
    {
      fill(255, 170, 10);
      stroke(255, 170, 10);
      if(interpolate)
      {
        ellipse(x + 8 + ((frameStart + frameLength)*zoom - scrollpos)*15, y + 22, 11, 11);
      }
      else
      {
        rect(x + 3 + ((frameStart + frameLength)*zoom - scrollpos)*15, y + 17, 11, 11);
      }
    }
    fill(255);
    stroke(0);
    if(interpolate)
    {
      ellipse(x + 8 + ((frameStart + frameLength)*zoom - scrollpos)*15, y + 22, 9, 9);    
    }
    else
    {
      rect(x + 4 + ((frameStart + frameLength)*zoom - scrollpos)*15, y + 18, 9, 9); 
    }
  }
}

class Vector
{
  public double x, y, z;

  public Vector(double x, double y, double z)
  {
    this.x = x;
    this.y = y;
    this.z = z;
  }

  public Vector copy()
  {
    return new Vector(x, y, z);
  }

  public float x() { 
    return (float)x; 
  }
  public float y() { 
    return (float)y; 
  }
  public float z() { 
    return (float)z; 
  }
}
